# -*- coding: utf-8 -*-
#
# This file is part of Inspirehep.
# Copyright (C) 2016 CERN.
#
# Inspirehep is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Inspirehep is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Inspirehep; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.


"""Test NOTHING global variable value."""

from __future__ import absolute_import, print_function


from json_merger.nothing import Nothing, NOTHING


def test_nothing_eq():
    assert NOTHING == NOTHING
    assert NOTHING == Nothing()
    assert NOTHING is not None
    # Force evaluation of the __eq__ method.
    assert (NOTHING == {}) is False
    assert (NOTHING == []) is False
    assert (NOTHING == 0) is False
    assert (NOTHING == '') is False


def test_nothing_neq():
    # Force evaluation of the __neq__ method.
    assert (NOTHING != NOTHING) is False
    assert (NOTHING != Nothing()) is False
    assert NOTHING != {}
    assert NOTHING != []
    assert NOTHING != 0
    assert NOTHING != ''


def test_nothing_repr():
    assert str(NOTHING) == 'NOTHING'
    assert repr(NOTHING) == 'NOTHING'


def test_nothing_bool():
    assert not NOTHING



# -*- coding: utf-8 -*-
#
# This file is part of Inspirehep.
# Copyright (C) 2016 CERN.
#
# Inspirehep is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Inspirehep is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Inspirehep; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

from __future__ import absolute_import, print_function

import json

from pyrsistent import freeze, thaw

from .utils import force_list


class ConflictType(object):
    """Types of Conflict.

    Attributes:
        REORDER: The list specified by the path might need to be reordered.

        MANUAL_MERGE: The triple specified as the conflict body needs to be
            manually merged and added to the conflict path.

        ADD_BACK_TO_HEAD: The object specified as the conflict body might
            need to be added back to the list specified in the conflict's path.

        SET_FIELD: The object specified as the conflict body needs to be
            added at the path specifed in the conflict object.

        REMOVE_FIELD: The value or object present at the path specified in
            the path conflict needs to be removed.
    """
    pass


_CONFLICTS = (
    'REORDER',
    'MANUAL_MERGE',
    'ADD_BACK_TO_HEAD',
    'SET_FIELD',
    'REMOVE_FIELD'
)
for conflict_type in _CONFLICTS:
    setattr(ConflictType, conflict_type, conflict_type)


class Conflict(tuple):
    """Immutable and Hashable representation of a conflict.

    Attributes:
        conflict_type: A :class:`json_merger.conflict.ConflictType` member.

        path: A tuple containing the path to the conflictual field.

        body: Optional value representing the body of the conflict.

    Note:
        Even if the conflict body can be any arbitrary object, this is saved
        internally as an immutable object so that a Conflict instance can be
        safely used in sets or as a dict key.
    """

    # Based on http://stackoverflow.com/a/4828108
    # Compatible with Python<=2.6

    def __new__(cls, conflict_type, path, body):
        if conflict_type not in _CONFLICTS:
            raise ValueError('Bad Conflict Type %s' % conflict_type)
        body = freeze(body)
        return tuple.__new__(cls, (conflict_type, path, body))

    conflict_type = property(lambda self: self[0])
    path = property(lambda self: self[1])
    body = property(lambda self: thaw(self[2]))

    def with_prefix(self, root_path):
        """Returns a new conflict with a prepended prefix as a path."""
        return Conflict(self.conflict_type, root_path + self.path, self.body)

    def to_json(self):
        """Deserializes conflict to a JSON object.

        It returns list of:
            `json-patch <https://tools.ietf.org/html/rfc6902>`_ format.

        - REORDER, SET_FIELD become "op": "replace"
        - MANUAL_MERGE, ADD_BACK_TO_HEAD become "op": "add"
        - Path becomes `json-pointer <https://tools.ietf.org/html/rfc6901>`_
        - Original conflict type is added to "$type"
        """
        # map ConflictType to json-patch operator
        path = self.path
        if self.conflict_type in ('REORDER', 'SET_FIELD'):
            op = 'replace'
        elif self.conflict_type in ('MANUAL_MERGE', 'ADD_BACK_TO_HEAD'):
            op = 'add'
            path += ('-',)
        elif self.conflict_type == 'REMOVE_FIELD':
            op = 'remove'
        else:
            raise ValueError(
                'Conflict Type %s can not be mapped to a json-patch operation'
                % conflict_type
            )

        # stringify path array
        json_pointer = '/' + '/'.join(str(el) for el in path)

        conflict_values = force_list(self.body)
        conflicts = []
        for value in conflict_values:
            if value is not None or self.conflict_type == 'REMOVE_FIELD':
                conflicts.append({
                    'path': json_pointer,
                    'op': op,
                    'value': value,
                    '$type': self.conflict_type
                })

        return json.dumps(conflicts)



# -*- coding: utf-8 -*-
#
# This file is part of Inspirehep.
# Copyright (C) 2016 CERN.
#
# Inspirehep is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Inspirehep is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Inspirehep; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

from __future__ import absolute_import, print_function

from collections import deque

import six

from .comparator import DefaultComparator
from .nothing import NOTHING
from .stats import ListMatchStats

FIRST = 'first'


class BeforeNodes(object):
    """Edge in the match graph."""

    def __init__(self, head_node=None, update_node=None):
        self.head_node = head_node
        self.update_node = update_node

    def __repr__(self):
        return 'BeforeNodes <head_node: {}, update_node: {}>'.format(
            self.head_node, self.update_node)


class ListMatchGraphBuilder(object):

    def __init__(self, root, head, update, sources,
                 comparator_cls=DefaultComparator):
        self.root = root
        self.head = head
        self.update = update
        self.sources = sources

        self.root_head_comparator = comparator_cls(self.root, self.head)
        self.root_update_comparator = comparator_cls(self.root, self.update)
        self.head_update_comparator = comparator_cls(self.head, self.update)

        # Keys are (target, source), values are comparator_instance and
        # the source list from which to search.
        self.comparators = {
            ('root', 'head'): (self.root_head_comparator, 'l2'),
            ('head', 'root'): (self.root_head_comparator, 'l1'),
            ('root', 'update'): (self.root_update_comparator, 'l2'),
            ('update', 'root'): (self.root_update_comparator, 'l1'),
            ('head', 'update'): (self.head_update_comparator, 'l2'),
            ('update', 'head'): (self.head_update_comparator, 'l1'),
        }

        self.node_data = {}
        self.graph = {}
        self.head_stats = ListMatchStats(head, root)
        self.update_stats = ListMatchStats(update, root)

        self._node_src_indices = {}
        self._head_idx_to_node = {}
        self._update_idx_to_node = {}
        self._dirty_nodes = set()

        self._next_node_id = 0

        self.multiple_match_choice_idx = set()
        self.multiple_match_choices = []

    def _new_node_id(self):
        node_id = self._next_node_id
        self._next_node_id += 1
        return node_id

    def _push_node(self, root_elem, head_elem, update_elem):
        root_idx, root_obj = root_elem
        head_idx, head_obj = head_elem
        update_idx, update_obj = update_elem

        node_id = self._new_node_id()
        self.node_data[node_id] = (root_obj, head_obj, update_obj)
        self._node_src_indices[node_id] = (root_idx, head_idx, update_idx)

        if head_idx >= 0:
            self._head_idx_to_node[head_idx] = node_id
        if update_idx >= 0:
            self._update_idx_to_node[update_idx] = node_id

    def _get_matches(self, source, source_idx, source_obj):
        other_two = {'head': ['root', 'update'],
                     'update': ['root', 'head'],
                     'root': ['head', 'update']}
        indices = {'root': {}, 'head': {}, 'update': {}}
        indices[source][source_idx] = source_obj

        # Start a BFS of matching elements.
        q = deque([(source, source_idx)])
        while q:
            curr_src, curr_idx = q.popleft()
            for target in other_two[curr_src]:
                comparator, cmp_list = self.comparators[(target, curr_src)]
                # cmp_list is either 'l1' or 'l2'
                # (the paremeter for the comparator class convetion)
                matches = comparator.get_matches(cmp_list, curr_idx)
                for target_idx, target_obj in matches:
                    if target_idx in indices[target]:
                        continue
                    indices[target][target_idx] = target_obj
                    q.append((target, target_idx))
        result = {}
        for lst, res_indices in indices.items():
            if not res_indices:
                result[lst] = [(-1, NOTHING)]
            else:
                result[lst] = sorted(res_indices.items())

        return result['root'], result['head'], result['update']

    def _add_matches(self, root_elems, head_elems, update_elems):
        matches = [(r, h, u)
                   for r in root_elems
                   for h in head_elems
                   for u in update_elems]
        if len(matches) == 1:
            self._push_node(*matches[0])
        else:
            self.multiple_match_choice_idx.update([(r[0], h[0], u[0])
                                                   for r, h, u in matches])

    def _populate_nodes(self):
        for idx, obj in enumerate(self.head):
            r_elems, h_elems, u_elems = self._get_matches('head', idx, obj)
            if 'head' in self.sources:
                self._add_matches(r_elems, h_elems, u_elems)
            if len(r_elems) == 1 and r_elems[0][0] >= 0:
                self.head_stats.add_root_match(idx, r_elems[0][0])

        for idx, obj in enumerate(self.update):
            r_elems, h_elems, u_elems = self._get_matches('update', idx, obj)
            # Only add the node to the graph only if not already added.
            if ('update' in self.sources and
                    idx not in self._update_idx_to_node):
                self._add_matches(r_elems, h_elems, u_elems)
            if len(r_elems) == 1 and r_elems[0][0] >= 0:
                self.update_stats.add_root_match(idx, r_elems[0][0])

        # Add stats from built nodes.
        for root_idx, head_idx, update_idx in self._node_src_indices.values():
            if head_idx >= 0:
                self.head_stats.move_to_result(head_idx)
            if update_idx >= 0:
                self.update_stats.move_to_result(update_idx)

        # Move the unique multiple match indices to conflicts.
        for r_idx, h_idx, u_idx in self.multiple_match_choice_idx:
            r_obj = self.root[r_idx] if r_idx >= 0 else None
            h_obj = self.head[h_idx] if h_idx >= 0 else None
            u_obj = self.update[u_idx] if u_idx >= 0 else None
            self.multiple_match_choices.append((r_obj, h_obj, u_obj))

    def _get_next_node(self, source, indices):
        if source not in self.sources:
            return None
        idx_to_node = {
            'head': self._head_idx_to_node,
            'update': self._update_idx_to_node
        }[source]
        for idx in indices:
            if idx in idx_to_node:
                return idx_to_node[idx]
        return None

    def build_graph(self):
        self._populate_nodes()

        # Link a dummy first node before the first element of the sources
        # lists.
        self.node_data[FIRST] = (NOTHING, NOTHING, NOTHING)
        self.graph[FIRST] = BeforeNodes()

        next_head_node = self._get_next_node('head', range(len(self.head)))
        next_update_node = self._get_next_node('update',
                                               range(len(self.update)))
        self.graph[FIRST].head_node = next_head_node
        self.graph[FIRST].update_node = next_update_node

        # Link any other nodes with the elements that come after them in their
        # source lists.
        for node_id, node_indices in six.iteritems(self._node_src_indices):
            root_idx, head_idx, update_idx = node_indices
            head_next_l = []
            update_next_l = []
            if head_idx >= 0:
                head_next_l = range(head_idx + 1, len(self.head))
            if update_idx >= 0:
                update_next_l = range(update_idx + 1, len(self.update))

            next_head_node = self._get_next_node('head', head_next_l)
            next_update_node = self._get_next_node('update', update_next_l)
            self.graph[node_id] = BeforeNodes(next_head_node, next_update_node)

        return self.graph, self.node_data


def _get_traversal(next_nodes, pick_first):
    if pick_first == 'head':
        return [next_nodes.update_node, next_nodes.head_node]
    else:
        return [next_nodes.head_node, next_nodes.update_node]


def toposort(graph, pick_first='head'):
    """Toplogically sorts a list match graph.

    Tries to perform a topological sort using as tiebreaker the pick_first
    argument. If the graph contains cycles, raise ValueError.
    """
    in_deg = {}
    for node, next_nodes in six.iteritems(graph):
        for next_node in [next_nodes.head_node, next_nodes.update_node]:
            if next_node is None:
                continue
            in_deg[next_node] = in_deg.get(next_node, 0) + 1

    stk = [FIRST]
    ordered = []
    visited = set()
    while stk:
        node = stk.pop()
        visited.add(node)
        if node != FIRST:
            ordered.append(node)
        traversal = _get_traversal(graph.get(node, BeforeNodes()), pick_first)
        for next_node in traversal:
            if next_node is None:
                continue
            if next_node in visited:
                raise ValueError('Graph has a cycle')

            in_deg[next_node] -= 1
            if in_deg[next_node] == 0:
                stk.append(next_node)

    # Nodes may not be walked because they don't reach in degree 0.
    if len(ordered) != len(graph) - 1:
        raise ValueError('Graph has a cycle')
    return ordered


def sort_cyclic_graph_best_effort(graph, pick_first='head'):
    """Fallback for cases in which the graph has cycles."""
    ordered = []
    visited = set()
    # Go first on the pick_first chain then go back again on the others
    # that were not visited. Given the way the graph is built both chains
    # will always contain all the elements.
    if pick_first == 'head':
        fst_attr, snd_attr = ('head_node', 'update_node')
    else:
        fst_attr, snd_attr = ('update_node', 'head_node')

    current = FIRST
    while current is not None:
        visited.add(current)
        current = getattr(graph[current], fst_attr)
        if current not in visited and current is not None:
            ordered.append(current)
    current = FIRST
    while current is not None:
        visited.add(current)
        current = getattr(graph[current], snd_attr)
        if current not in visited and current is not None:
            ordered.append(current)
    return ordered



# -*- coding: utf-8 -*-
#
# This file is part of Inspirehep.
# Copyright (C) 2016 CERN.
#
# Inspirehep is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Inspirehep is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Inspirehep; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.


"""Test merger corner cases that are unlikely to appear as usage scenarios."""

from __future__ import absolute_import, print_function

from json_merger.comparator import PrimaryKeyComparator


def test_multiple_primary_keys():
    class MyComp(PrimaryKeyComparator):
        primary_key_fields = ['id', 'f.id']

    lst = [{'id': 0}, {'id': 1}, {'f': {'id': 0}}, {'f': {'id': 1}}]
    inst = MyComp(lst, lst)

    for i, obj in enumerate(lst):
        assert inst.get_matches('l1', i) == [(i, obj)]
        assert inst.get_matches('l2', i) == [(i, obj)]


def test_list_of_primary_keys():
    class MyComp(PrimaryKeyComparator):
        primary_key_fields = [['id1', 'id2']]

    lst1 = [{'id1': 0, 'data': 1},
            {'id2': 0, 'data': 1},
            {'id1': 1, 'id2': 1, 'data': 1}]
    lst2 = [{'id1': 0, 'data': 2},
            {'id2': 0, 'data': 2},
            {'id1': 1, 'id2': 1, 'data': 2},
            {'id1': 1, 'id2': 0, 'data': 2}]

    inst = MyComp(lst1, lst2)

    assert not inst.get_matches('l1', 0)
    assert not inst.get_matches('l1', 1)
    assert not inst.get_matches('l2', 0)
    assert not inst.get_matches('l2', 1)
    assert not inst.get_matches('l2', 3)

    assert inst.get_matches('l1', 2) == [(2, lst2[2])]
    assert inst.get_matches('l2', 2) == [(2, lst1[2])]


def test_list_of_primary_keys_normalization():
    class MyComp(PrimaryKeyComparator):
        primary_key_fields = [['id1', 'id2']]
        normalization_functions = {'id2': str.lower}

    lst1 = [{'id1': 0, 'data': 1},
            {'id2': 'a', 'data': 1},
            {'id1': 1, 'id2': 'a', 'data': 1}]
    lst2 = [{'id1': 0, 'data': 2},
            {'id2': 'A', 'data': 2},
            {'id1': 1, 'id2': 'A', 'data': 2},
            {'id1': 1, 'id2': 'B', 'data': 2}]

    inst = MyComp(lst1, lst2)

    assert not inst.get_matches('l1', 0)
    assert not inst.get_matches('l1', 1)
    assert not inst.get_matches('l2', 0)
    assert not inst.get_matches('l2', 1)
    assert not inst.get_matches('l2', 3)

    assert inst.get_matches('l1', 2) == [(2, lst2[2])]
    assert inst.get_matches('l2', 2) == [(2, lst1[2])]



# -*- coding: utf-8 -*-
#
# This file is part of Inspirehep.
# Copyright (C) 2016 CERN.
#
# Inspirehep is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Inspirehep is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Inspirehep; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

from __future__ import absolute_import, print_function


class ListMatchStats(object):
    """Class for holding list entity matching stats."""

    def __init__(self, lst, root):
        """
        Args:
            lst: The list of elements that needs to be matched.
            root: The ancestor of the list of elements that needs to be
                  matched.

        Attributes:
            in_result_idx: Indices of elements in lst that are present in the
                end result.

            in_result: Elements in lst that are present in the end result.

            not_in_result_idx: Indices of elements in lst that are not present
                in the end result.

            not_in_result: Elements in lst that are not present in the end
                result.

            not_in_result_root_match_idx: Indices of elements that are not in
                the end result but were matched with root elements.

            not_in_result_root_match: Elements that are not in the end result
                but were matched with root elements.

            not_in_result_not_root_match_idx: Indices of elements that are not
                in the end result and were not matched with any root elements.

            not_in_result_not_root_match: Elements that are not in the end
                result and were not matched with any root elements.

            not_in_result_root_match_pairs: Pairs of (lst, root) elements
                that are not in the end result but were matched.
        """
        self.lst = lst
        self.root = root

        self.in_result_idx = set()
        self.not_in_result_root_match_idx = set()
        self.root_matches = {}

    def move_to_result(self, lst_idx):
        """Moves element from lst available at lst_idx."""
        self.in_result_idx.add(lst_idx)

        if lst_idx in self.not_in_result_root_match_idx:
            self.not_in_result_root_match_idx.remove(lst_idx)

    def add_root_match(self, lst_idx, root_idx):
        """Adds a match for the elements avaialble at lst_idx and root_idx."""
        self.root_matches[lst_idx] = root_idx
        if lst_idx in self.in_result_idx:
            return

        self.not_in_result_root_match_idx.add(lst_idx)

    @property
    def not_in_result_idx(self):
        return set(range(len(self.lst))).difference(self.in_result_idx)

    @property
    def not_in_result_not_root_match_idx(self):
        return self.not_in_result_idx.difference(
            self.not_in_result_root_match_idx)

    @property
    def in_result(self):
        return [self.lst[e] for e in self.in_result_idx]

    @property
    def not_in_result(self):
        return [self.lst[e] for e in self.not_in_result_idx]

    @property
    def not_in_result_root_match(self):
        return [self.lst[e] for e in self.not_in_result_root_match_idx]

    @property
    def not_in_result_not_root_match(self):
        return [self.lst[e] for e in self.not_in_result_not_root_match_idx]

    @property
    def not_in_result_root_match_pairs(self):
        return [(self.lst[e], self.root[self.root_matches[e]])
                for e in self.not_in_result_root_match_idx]

    @property
    def not_matched_root_objects(self):
        matched_root_idx = set(self.root_matches.values())
        return [o for idx, o in enumerate(self.root)
                if idx not in matched_root_idx]



# -*- coding: utf-8 -*-
#
# This file is part of Inspirehep.
# Copyright (C) 2016 CERN.
#
# Inspirehep is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Inspirehep is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Inspirehep; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

"""Test dictdiffer wrappers."""

from __future__ import absolute_import, print_function

import pytest
import json

from json_merger.conflict import Conflict


def test_to_json_with_reorder():
    conflict = Conflict('REORDER', ('foo', 'bar'), {})
    conflict_json = conflict.to_json()
    assert json.loads(conflict_json) == [
        {
            '$type': 'REORDER',
            'op': 'replace',
            'path': '/foo/bar',
            'value': {}
        }
    ]


def test_to_json_with_set_field():
    conflict = Conflict('SET_FIELD', ('foo', 'bar'), {})
    conflict_json = conflict.to_json()
    assert json.loads(conflict_json) == [
        {
            '$type': 'SET_FIELD',
            'op': 'replace',
            'path': '/foo/bar',
            'value': {}
        }
    ]


def test_to_json_with_manual_merge():
    body = [
        None,
        {'foo1': 'bar1'},
        {'foo2': 'bar2'}
    ]
    conflict = Conflict('MANUAL_MERGE', ('foo', 'bar'), body)
    conflict_json = conflict.to_json()
    assert json.loads(conflict_json) == [
        {
            '$type': 'MANUAL_MERGE',
            'op': 'add',
            'path': '/foo/bar/-',
            'value': {'foo1': 'bar1'}
        },
        {
            '$type': 'MANUAL_MERGE',
            'op': 'add',
            'path': '/foo/bar/-',
            'value': {'foo2': 'bar2'}
        }
    ]


def test_to_json_with_add_back_to_head():
    conflict = Conflict('ADD_BACK_TO_HEAD', ('foo', 'bar'), {})
    conflict_json = conflict.to_json()
    assert json.loads(conflict_json) == [
        {
            '$type': 'ADD_BACK_TO_HEAD',
            'op': 'add',
            'path': '/foo/bar/-',
            'value': {}
        }
    ]


def test_to_json_with_remove_field():
    conflict = Conflict('REMOVE_FIELD', ('foo', 'bar'), None)
    conflict_json = conflict.to_json()
    assert json.loads(conflict_json) == [
        {
            '$type': 'REMOVE_FIELD',
            'op': 'remove',
            'path': '/foo/bar',
            'value': None
        }
    ]


def test_to_json_when_path_has_integers():
    conflict = Conflict('REMOVE_FIELD', ('foo', 0, 'bar', 1), None)
    conflict_json = conflict.to_json()
    assert json.loads(conflict_json) == [
        {
            '$type': 'REMOVE_FIELD',
            'op': 'remove',
            'path': '/foo/0/bar/1',
            'value': None
        }
    ]



# -*- coding: utf-8 -*-
#
# This file is part of Inspirehep.
# Copyright (C) 2016 CERN.
#
# Inspirehep is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Inspirehep is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Inspirehep; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

from __future__ import absolute_import, print_function

from .nothing import NOTHING
from .utils import get_obj_at_key_path


class BaseComparator(object):
    """Abstract base class for Entity Comparison."""

    def __init__(self, l1, l2):
        """
        Args:
            l1: First list of entities.
            l2: Second list of entities.
        """
        self.l1 = l1
        self.l2 = l2
        self.matches = set()
        self.process_lists()

    def process_lists(self):
        """Do any preprocessing of the lists."""
        for l1_idx, obj1 in enumerate(self.l1):
            for l2_idx, obj2 in enumerate(self.l2):
                if self.equal(obj1, obj2):
                    self.matches.add((l1_idx, l2_idx))

    def equal(self, obj1, obj2):
        """Implementation of object equality."""
        raise NotImplementedError()

    def get_matches(self, src, src_idx):
        """Get elements equal to the idx'th in src from the other list.

        e.g. get_matches(self, 'l1', 0) will return all elements from self.l2
        matching with self.l1[0]
        """
        if src not in ('l1', 'l2'):
            raise ValueError('Must have one of "l1" or "l2" as src')
        if src == 'l1':
            target_list = self.l2
        else:
            target_list = self.l1
        comparator = {
            'l1': lambda s_idx, t_idx: (s_idx, t_idx) in self.matches,
            'l2': lambda s_idx, t_idx: (t_idx, s_idx) in self.matches,
        }[src]

        return [(trg_idx, obj) for trg_idx, obj in enumerate(target_list)
                if comparator(src_idx, trg_idx)]


class PrimaryKeyComparator(BaseComparator):
    """Considers two objects as equal if they have the same primary key.

    If two objects have at least one of the configured primary_key_fields equal
    then they are equal. A primary key field can be any of:

    string: Two objects are equal if the values at the given key paths
            are equal. Example:
                For 'key1.key2' the objects are equal if
                obj1['key1']['key2'] == obj2['key1']['key2'].
    list: Two objects are equal if all the values at the key paths
          in the list are equal. Example:
                For ['key1', 'key2.key3'] the objects are equal if
                obj1['key1'] == obj2['key1'] and
                obj1['key2']['key3'] == obj2['key2']['key3'].

    For normalizing the fields in the objects to be compared, one can add
    a normalization function for each field in the normalization_functions
    dict.

    Example:
        Setting the normalization_functions field to:
            ``{'key1': str.lower}``
        would normalize
            obj1 = {'key1': 'ID123'} and obj2 = {'key1': 'id123'} to
            obj1 = {'key1': 'id123'} and obj2 = {'key1': 'id123'}
    """

    primary_key_fields = ['pk']
    normalization_functions = {}

    def _have_field_equal(self, obj1, obj2, field):
        key_path = tuple(k for k in field.split('.') if k)
        o1 = get_obj_at_key_path(obj1, key_path, NOTHING)
        o2 = get_obj_at_key_path(obj2, key_path, NOTHING)
        if o1 == NOTHING or o2 == NOTHING:
            return False

        fn = self.normalization_functions.get(field, lambda x: x)
        return fn(o1) == fn(o2)

    def equal(self, obj1, obj2):
        if obj1 == obj2:
            return True

        for field_set in self.primary_key_fields:
            if not isinstance(field_set, list):
                field_set = [field_set]
            checks = [self._have_field_equal(obj1, obj2, field)
                      for field in field_set]
            if all(checks):
                return True

        return False


class DefaultComparator(BaseComparator):
    """Two objects are the same entity if they are fully equal."""

    def equal(self, obj1, obj2):
        return obj1 == obj2



# -*- coding: utf-8 -*-
#
# This file is part of Inspirehep.
# Copyright (C) 2016 CERN.
#
# Inspirehep is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Inspirehep is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Inspirehep; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.


"""Test list aligner correct output."""

from __future__ import absolute_import, print_function

import pytest

from json_merger.config import UnifierOps
from json_merger.conflict import ConflictType
from json_merger.comparator import PrimaryKeyComparator
from json_merger.errors import MergeError
from json_merger.list_unify import ListUnifier
from json_merger.nothing import NOTHING


def test_value_error():
    with pytest.raises(ValueError):
        ListUnifier([], [], [], 'BAD_OPERATION')


def test_keep_only_head_entitites():
    root = [1, 2]
    head = [1, 2, 3]
    update = [6, 5, 4, 3, 2]

    u = ListUnifier(root, head, update, UnifierOps.KEEP_ONLY_HEAD_ENTITIES)
    u.unify()

    assert u.unified == [(1, 1, NOTHING), (2, 2, 2), (NOTHING, 3, 3)]


def test_keep_only_update_entities():
    root = [1, 2]
    head = [6, 5, 4, 3, 2]
    update = [1, 2, 3]

    u = ListUnifier(root, head, update, UnifierOps.KEEP_ONLY_UPDATE_ENTITIES)
    u.unify()

    assert u.unified == [(1, NOTHING, 1), (2, 2, 2), (NOTHING, 3, 3)]


def test_keep_update_and_head_ent_head_fst():
    root = [1, 2]
    head = [5, 4, 3, 2]
    update = [10, 3, 1, 2, 11]

    u = ListUnifier(root, head, update,
                    UnifierOps.KEEP_UPDATE_AND_HEAD_ENTITIES_HEAD_FIRST)
    u.unify()

    assert u.unified == [(NOTHING, 5, NOTHING), (NOTHING, 4, NOTHING),
                         (NOTHING, NOTHING, 10),
                         (NOTHING, 3, 3), (1, NOTHING, 1),
                         (2, 2, 2),
                         (NOTHING, NOTHING, 11)]


def test_keep_update_and_head_ent_update_fst():
    root = [1, 2]
    head = [5, 4, 3, 2]
    update = [10, 3, 1, 2, 11]

    u = ListUnifier(root, head, update,
                    UnifierOps.KEEP_UPDATE_AND_HEAD_ENTITIES_UPDATE_FIRST)
    u.unify()

    assert u.unified == [(NOTHING, NOTHING, 10),
                         (NOTHING, 5, NOTHING), (NOTHING, 4, NOTHING),
                         (NOTHING, 3, 3), (1, NOTHING, 1),
                         (2, 2, 2),
                         (NOTHING, NOTHING, 11)]


def test_keep_update_and_head_ent_head_fst_fallback():
    root = [1, 2]
    head = [1, 2, 3]
    update = [7, 3, 6, 1, 5, 2, 4]

    u = ListUnifier(root, head, update,
                    UnifierOps.KEEP_UPDATE_AND_HEAD_ENTITIES_HEAD_FIRST)

    with pytest.raises(MergeError) as excinfo:
        u.unify()

    assert len(excinfo.value.content) == 1
    conflict = excinfo.value.content[0]
    assert conflict.conflict_type == ConflictType.REORDER
    assert conflict.path == ()
    assert conflict.body is None

    assert u.unified == [(1, 1, 1), (2, 2, 2), (NOTHING, 3, 3),
                         (NOTHING, NOTHING, 7),
                         (NOTHING, NOTHING, 6),
                         (NOTHING, NOTHING, 5),
                         (NOTHING, NOTHING, 4)]


def test_error_on_head_delete():
    root = [1, 2]
    head = [1, 2, 3, 5]
    update = [1, 2, 4]

    u = ListUnifier(root, head, update,
                    UnifierOps.KEEP_UPDATE_ENTITIES_CONFLICT_ON_HEAD_DELETE)

    with pytest.raises(MergeError) as excinfo:
        u.unify()
    assert len(excinfo.value.content) == 2

    expect_removed = {3, 5}
    removed = set()
    for conflict in excinfo.value.content:
        assert conflict.conflict_type == ConflictType.ADD_BACK_TO_HEAD
        assert conflict.path == ()
        removed.add(conflict.body)
    assert removed == expect_removed

    assert u.unified == [(1, 1, 1), (2, 2, 2), (NOTHING, NOTHING, 4)]


def test_no_error_on_head_delete_from_root():
    root = [1, 2]
    head = [1, 2]
    update = [1, 4]

    u = ListUnifier(root, head, update,
                    UnifierOps.KEEP_UPDATE_ENTITIES_CONFLICT_ON_HEAD_DELETE)
    u.unify()
    assert u.unified == [(1, 1, 1), (NOTHING, NOTHING, 4)]


def test_error_on_multiple_match():
    root = [2, 3]
    head = [1, 1, 2, 3, 3]
    update = [1, 2, 3]

    u = ListUnifier(root, head, update,
                    UnifierOps.KEEP_ONLY_UPDATE_ENTITIES)
    with pytest.raises(MergeError) as excinfo:
        u.unify()

    assert len(excinfo.value.content) == 4
    expected_bodies = [(None, 1, 1), (None, 1, 1), (3, 3, 3), (3, 3, 3)]
    for conflict in excinfo.value.content:
        assert conflict.conflict_type == ConflictType.MANUAL_MERGE
        assert conflict.path == ()
        assert conflict.body in expected_bodies
        expected_bodies.remove(conflict.body)
    assert not expected_bodies

    assert u.unified == [(2, 2, 2)]


def test_multiple_match_symmetry():
    root = []
    l1 = [1, 2, 3, 3, 3]
    l2 = [1, 2, 3]

    u1 = ListUnifier(root, l1, l2,
                     UnifierOps.KEEP_UPDATE_AND_HEAD_ENTITIES_HEAD_FIRST)
    u2 = ListUnifier(root, l2, l1,
                     UnifierOps.KEEP_UPDATE_AND_HEAD_ENTITIES_UPDATE_FIRST)

    with pytest.raises(MergeError) as u1_excinfo:
        u1.unify()
    with pytest.raises(MergeError) as u2_excinfo:
        u2.unify()

    assert len(u1_excinfo.value.content) == 3
    assert len(u2_excinfo.value.content) == 3

    for conflict in u1_excinfo.value.content + u2_excinfo.value.content:
        assert conflict.conflict_type == ConflictType.MANUAL_MERGE
        assert conflict.path == ()
        assert conflict.body == (None, 3, 3)

    assert u1.unified == [(NOTHING, 1, 1), (NOTHING, 2, 2)]
    assert u2.unified == [(NOTHING, 1, 1), (NOTHING, 2, 2)]


def test_stats():
    root = [1, 2, 10]
    head = [1, 3, 4, 2]
    update = [1, 3, 5]

    u = ListUnifier(root, head, update,
                    UnifierOps.KEEP_ONLY_UPDATE_ENTITIES)
    u.unify()

    assert sorted(u.head_stats.in_result) == [1, 3]
    assert sorted(u.head_stats.not_in_result) == [2, 4]
    assert sorted(u.head_stats.not_in_result_root_match) == [2]
    assert sorted(u.head_stats.not_in_result_root_match_idx) == [3]
    assert sorted(u.head_stats.not_in_result_not_root_match) == [4]
    assert sorted(u.head_stats.not_in_result_root_match_pairs) == [(2, 2)]
    assert sorted(u.head_stats.not_matched_root_objects) == [10]

    assert sorted(u.update_stats.in_result) == [1, 3, 5]
    assert sorted(u.update_stats.not_in_result) == []
    assert sorted(u.update_stats.not_in_result_root_match) == []
    assert sorted(u.update_stats.not_in_result_not_root_match) == []
    assert sorted(u.update_stats.not_matched_root_objects) == [2, 10]


def test_transitive_equality():
    class Comp(PrimaryKeyComparator):
        primary_key_fields = ['id0', 'id1']

    only0 = {'id0': 0}
    only1 = {'id1': 1}
    both = {'id0': 0, 'id1': 1}

    u = ListUnifier([only0], [both], [only1],
                    UnifierOps.KEEP_ONLY_UPDATE_ENTITIES, Comp)
    u.unify()
    assert u.unified == [(only0, both, only1)]

    u = ListUnifier([only0], [only1], [both],
                    UnifierOps.KEEP_ONLY_HEAD_ENTITIES, Comp)
    u.unify()
    assert u.unified == [(only0, only1, both)]

    u = ListUnifier([only0], [only1], [both],
                    UnifierOps.KEEP_UPDATE_AND_HEAD_ENTITIES_HEAD_FIRST, Comp)
    u.unify()
    assert u.unified == [(only0, only1, both)]



# -*- coding: utf-8 -*-
#
# This file is part of Inspirehep.
# Copyright (C) 2016 CERN.
#
# Inspirehep is free software; you can redistribute it
# and/or modify it under the terms of the GNU General Public License as
# published by the Free Software Foundation; either version 2 of the
# License, or (at your option) any later version.
#
# Inspirehep is distributed in the hope that it will be
# useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Inspirehep; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
# MA 02111-1307, USA.
#
# In applying this license, CERN does not
# waive the privileges and immunities granted to it by virtue of its status
# as an Intergovernmental Organization or submit itself to any jurisdiction.

from __future__ import absolute_import, print_function

from .nothing import NOTHING


def get_obj_at_key_path(obj, key_path, default=None):
    current = obj
    for k in key_path:
        try:
            current = current[k]
        except (KeyError, IndexError, TypeError):
            return default
    return current


def set_obj_at_key_path(obj, key_path, value, raise_key_error=True):
    try:
        return _set_obj_at_key_path(obj, key_path, value)
    except KeyError as e:
        if raise_key_error:
            raise e
        else:
            return obj


def _set_obj_at_key_path(obj, key_path, value):
    # We are setting the obj param to another value.
    if len(key_path) == 0:
        return value

    # Try to get the parent of the object to be set.
    parent = get_obj_at_key_path(obj, key_path[:-1], NOTHING)
    if parent == NOTHING:
        raise KeyError(key_path)
    try:
        parent[key_path[-1]] = value
    except (KeyError, IndexError, TypeError):
        raise KeyError(key_path)
    return obj


def del_obj_at_key_path(obj, key_path, raise_key_error=True):
    obj = get_obj_at_key_path(obj, key_path[:-1], NOTHING)
    if obj != NOTHING:
        try:
            del obj[key_path[-1]]
        except (KeyError, IndexError, TypeError):
            if raise_key_error:
                raise KeyError(key_path)
    elif raise_key_error:
        raise KeyError(key_path)


def has_prefix(key_path, prefix):
    return len(prefix) <= len(key_path) and key_path[:len(prefix)] == prefix


def remove_prefix(key_path, prefix):
    if not has_prefix(key_path, prefix):
        raise ValueError('Bad Prefix {}'.format(prefix))
    return key_path[len(prefix):]


def get_dotted_key_path(key_path, filter_int_keys=False):
    return '.'.join(k for k in key_path
                    if not isinstance(k, int) and filter_int_keys)


def get_conf_set_for_key_path(conf_set, key_path):
    prefix = get_dotted_key_path(key_path, True)
    return set(remove_prefix(k, prefix).lstrip('.')
               for k in conf_set if has_prefix(k, prefix))


def force_list(data):
    if not isinstance(data, (list, tuple, set)):
        return [data]
    elif isinstance(data, (tuple, set)):
        return list(data)
    return data


def dedupe_list(l):
    """Remove duplicates from a list preserving the order.

    We might be tempted to use the list(set(l)) idiom, but it doesn't preserve
    the order, which hinders testability and does not work for lists with
    unhashable elements.
    """
    result = []

    for el in l:
        if el not in result:
            result.append(el)

    return result
